from flask_wtf import FlaskForm
from wtforms import BooleanField, HiddenField, PasswordField, SelectField, StringField, SubmitField
from wtforms.validators import DataRequired


class RoomCreate(FlaskForm):
    room_name = StringField('Room name', validators=[DataRequired()])
    password = PasswordField('Room password (optional)')
    public = BooleanField('Public (i.e. listed to the right)')
    guest_limit = SelectField('Guest limit', coerce=int,
                               choices=[(2, '2'), (3, '3'), (4, '4'), (5, '5'), (0, 'No limit')])
    submit = SubmitField('Create')


class RoomJoin(FlaskForm):
    room_id = HiddenField()
    password = PasswordField('Password: ')
    submit = SubmitField('Join')



import asyncio
from time import time

class MTimer:
    def __init__(self, timeout, callback, **kwargs):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.create_task(self._run())
        self._kwargs = kwargs

    async def _run(self):
        await asyncio.sleep(self._timeout)
        await self._callback(**self._kwargs)

    def cancel(self):
        self._task.cancel()


def time_ms():
    return int(time() * 1000)



from logging.config import dictConfig

dictConfig({
    'version': 1,
    'formatters': {'default': {
        #'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        'format': '[%(levelname)s] in %(module)s %(funcName)s %(lineno)d: %(message)s',
    }},
    'handlers': {'wsgi': {
        'class': 'logging.StreamHandler',
        'formatter': 'default'
    }},
    'root': {
        'level': 'INFO',
        'handlers': ['wsgi']
    }
})

import quart.flask_patch

from quart import Quart
from flask_bootstrap import Bootstrap

from .config import Config

app = Quart(__name__)
app.config.from_object(Config)
bootstrap = Bootstrap(app)

from camus import routes



import uuid

import pytest

from camus.chat import ChatClient, ChatManager, ChatRoom, get_chat_manager

pytestmark = pytest.mark.asyncio

def test_get_chat_manager():
    assert get_chat_manager() is not None

class TestChatRoom:
    async def test_add_client(self):
        client_id = uuid.uuid4().hex
        client = ChatClient(client_id)
        room = ChatRoom('junk')
        room.add_client(client)
        assert client_id in room.clients
        assert client in room.clients.values()

    async def test_remove_client(self):
        client_id = uuid.uuid4().hex
        client = ChatClient(client_id)
        room = ChatRoom('garbage')
        room.add_client(client)
        room.remove_client(client)
        assert client_id not in room.clients
        assert client not in room.clients.values()

    async def test_get_clients(self):
        client1 = ChatClient(uuid.uuid4().hex)
        client2 = ChatClient(uuid.uuid4().hex)
        room = ChatRoom('trash')
        room.add_client(client1)
        room.add_client(client2)
        clients = room.get_clients()
        assert client1 in clients
        assert client2 in clients

    async def test_info(self):
        client_id = uuid.uuid4().hex
        client_username = 'Mr. Bill'
        client = ChatClient(client_id, username=client_username)
        room_id = 'sludge'
        room = ChatRoom(room_id)
        room.add_client(client)
        room_info = room.info
        assert 'room_id' in room_info
        assert 'clients' in room_info
        assert room_info['room_id'] == room_id
        assert client_id in [item['id'] for item in room_info['clients']]
        assert client_username in [item['username'] for item in room_info['clients']]

    async def test_is_full(self):
        client1 = ChatClient(uuid.uuid4().hex)
        client2 = ChatClient(uuid.uuid4().hex)
        room = ChatRoom('clutter', guest_limit=2)
        room.add_client(client1)
        room.add_client(client2)
        assert room.is_full()

        room = ChatRoom('debris', guest_limit=None)
        room.add_client(client1)
        room.add_client(client2)
        assert not room.is_full()

class TestChatManager:
    async def test_add_room(self):
        manager = get_chat_manager()
        room = ChatRoom('waste')
        manager.add_room(room)
        assert room.id in manager.rooms

    async def test_create_client(self):
        manager = get_chat_manager()
        client = manager.create_client()
        assert client is not None
        assert client.id is not None

        client_id = uuid.uuid4().hex
        client = manager.create_client(client_id)
        assert client is not None
        assert client.id == client_id

    async def test_create_room(self):
        manager = get_chat_manager()
        room_id = uuid.uuid4().hex
        room = manager.create_room(room_id)
        assert room is not None
        assert room.id == room_id
        assert room_id in manager.rooms
        assert manager.rooms[room_id] == room

    async def test_remove_client(self):
        manager = get_chat_manager()
        client = manager.create_client()
        room = manager.create_room('rubbish')
        room.add_client(client)
        await manager.remove_client(client)
        assert client.id not in manager.clients
        assert client not in manager.clients.values()

    async def test_get_room(self):
        manager = get_chat_manager()
        room = manager.create_room('refuse')
        got_room = manager.get_room(room.id)
        assert room == got_room

    async def test_get_public_rooms(self):
        manager = get_chat_manager()
        public_room = manager.create_room('public', is_public=True)
        assert public_room in manager.get_public_rooms()

        private_room = manager.create_room('private', is_public=False)
        assert private_room not in manager.get_public_rooms()



import asyncio
import hmac
import json
import logging
import uuid

from base64 import b64encode
from time import time

from pyee import AsyncIOEventEmitter
from slugify import slugify

from werkzeug.security import generate_password_hash, check_password_hash

from camus import app
from camus.util import MTimer, time_ms


_chat_manager = None


def get_chat_manager():
    global _chat_manager
    if _chat_manager is None:
        _chat_manager = ChatManager()
    return _chat_manager


def generate_turn_creds(key, client_id):
    expiration = int(time()) + 6 * 60 * 60  # creds expire after 6 hrs
    username = '{}:{}'.format(expiration, client_id)
    token = hmac.new(key.encode(), msg=username.encode(), digestmod='SHA1')
    password = b64encode(token.digest()).decode()

    return username, password


class ChatException(Exception):
    pass


class ChatRoom(AsyncIOEventEmitter):
    def __init__(self, name, password=None, guest_limit=None, admin_list=None, is_public=False):
        super().__init__()

        self.name = name
        self.id = slugify(name)
        self.clients = {}
        self.password_hash = None if password is None else generate_password_hash(password)
        self.guest_limit = guest_limit
        self.admin_list = admin_list if admin_list is not None else []
        self.is_public = is_public
        self._last_active = time_ms()

        async def check_expire():
            now = time_ms() / 1000
            last_active = self.last_active / 1000

            if now - last_active > self._reap_timeout:
                logging.info('Room %s expired', self.id)
                self.emit('expire')
            else:
                new_timeout = last_active + self._reap_timeout - now
                self._timer = MTimer(new_timeout, check_expire)

        self._reap_timeout = 3600
        self._timer = MTimer(self._reap_timeout, check_expire)

        logging.info('Created room %s', self.id)


    @property
    def last_active(self):
        last_seen = [client.last_seen for client in self.clients.values()]
        self._last_active = max([self._last_active, *last_seen])
        return self._last_active

    @property
    def active_ago(self):
        return int((time_ms() - self.last_active) / 60000)

    @property
    def info(self):
        clients = [{'id': client.id, 'username': client.username}
                   for client in self.get_clients()]

        return {'room_id': self.id, 'clients': clients}

    def authenticate(self, password=None):
        if password is None:
            return self.password_hash is None

        return check_password_hash(self.password_hash, password)

    def is_full(self):
        return self.guest_limit is not None and len(self.clients) == self.guest_limit

    def add_client(self, client):
        if self.is_full():
            raise ChatException('Guest limit already reached')

        self.clients[client.id] = client
        client.room = self

    def remove_client(self, client):
        logging.info('Removing client {} from room {}'.format(client.id, self.id))
        self._last_active = max(self._last_active, client.last_seen)
        client.room = None
        self.clients.pop(client.id, None)
        logging.info('{} clients remaining in room {}'.format(len(self.clients), self.id))

    def get_clients(self):
        return self.clients.values()

    def broadcast(self, message):
        for client in self.get_clients():
            message.receiver = client.id
            client.send(message.json())

    async def shutdown(self):
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

        for client in self.clients.values():
            self.remove_client(client)
            await client.shutdown()


class ChatClient(AsyncIOEventEmitter):
    def __init__(self, id, username=None, room=None, is_admin=False):
        super().__init__()

        self.id = id
        self.username = username if username is not None else 'Major Tom'
        self.room = room
        self.is_admin = is_admin
        self.inbox = asyncio.Queue()
        self.outbox = asyncio.Queue()
        self._inbox_task = asyncio.create_task(
            self._process_inbox()
        )

        # Used by ChatManager for reaping
        # TODO: there is probably a cleaner solution
        self.last_seen = time_ms()
        self.timer = None

        logging.info('Created client %s', self.id)

    async def _process_inbox(self):
        while True:
            message = await self.inbox.get()
            self.last_seen = time_ms()
            self.emit('message', message)

    def send(self, data):
        try:
            self.outbox.put_nowait(data)
            message_type = json.loads(data)['type']
            logging.info('Sent %s to client %s', message_type, self.id)

        except Exception as e:
            logging.info('Couldn\'t send message to client {}: {}'
                         .format(self.id, e))

    def ping(self):
        message = ChatMessage()
        message.sender = 'ground control'
        message.receiver = self.id
        message.type = 'ping'
        message.data = time_ms()
        self.send(message.json())

    async def shutdown(self):
        message = ChatMessage()
        message.sender = 'ground control'
        message.receiver = self.id
        message.type = 'bye'
        message.data = time_ms()
        self.send(message.json())

        if self.timer is not None:
            self.timer.cancel()
            self.timer = None

        if self._inbox_task is not None:
            self._inbox_task.cancel()
            self._inbox_task = None

        logging.info('Shut down client %s', self.id)

class ChatMessage:
    def __init__(self, message=None):
        if isinstance(message, str):
            _json = json.loads(message)
        elif message is None:
            _json = {}
        else:
            _json = message

        # TODO: verify the sender against the client connection
        self.sender = _json.get('sender')
        self.receiver = _json.get('receiver')
        self.type = _json.get('type')
        self.data = _json.get('data')

    def json(self):
        return json.dumps(self.__dict__)


class ChatManager:
    def __init__(self):
        self.rooms = {}
        self._message_address = "ground control"
        self._reap_timeout = 60

    @property
    def clients(self):
        return {client.id: client for room in self.rooms.values()
                for client in room.clients.values()}

    async def _handle_message(self, message, client):
        chat_message = self._parse_message(message, client)

        if chat_message.receiver == self._message_address:
            await self._handle_local_message(chat_message, client)
            return

        if chat_message.receiver == 'room':
            self._handle_room_message(chat_message, client)
            return

        if chat_message.receiver not in self.clients:
            logging.info('Message recipient %s does not exist', chat_message.receiver)
            # TODO: reply with error
            return

        to_client = self.clients[chat_message.receiver]
        to_client.send(chat_message.json())

    async def _handle_local_message(self, message, client):
        reply = ChatMessage()
        reply.sender = self._message_address
        reply.receiver = client.id

        if message.type == 'ping':
            reply.type = 'pong'
            reply.data = message.data
        elif message.type == 'pong':
            logging.info('Got pong {} from client {}'.format(message.data, message.sender))
            return
        elif message.type == 'profile':
            username = message.data.get('username')
            if username:
                client.username = username
            self.broadcast_room_info(client.room)
            return
        elif message.type == 'get-room-info':
            reply.type = 'room-info'
            reply.data = client.room.info
        elif message.type == 'get-ice-servers':
            reply.type = 'ice-servers'
            reply.data = self.get_ice_servers(client.id)
        elif message.type == 'greeting':
            logging.info('Greeting received from client {}: {}'.format(message.sender, message.data))
            return
        elif message.type == 'bye':
            await self.remove_client(client)
            return
        else:
            reply.type = 'error'
            reply.data = 'Unknown message type: {}'.format(message.type)

        client.send(reply.json())

    def _handle_room_message(self, message, client):
        room = client.room
        for c in room.clients.values():
            c.send(message.json())

    def _parse_message(self, message, client):
        chat_message = ChatMessage(message)
        if chat_message.sender is None:
            chat_message.sender = client.id
        return chat_message

    def broadcast_room_info(self, room):
        message = ChatMessage()
        message.sender = self._message_address
        message.type = 'room-info'
        message.data = room.info
        room.broadcast(message)

    def add_room(self, room):
        self.rooms[room.id] = room

    async def remove_client(self, client):
        room = client.room
        if room:
            room.remove_client(client)
            self.broadcast_room_info(room)

        await client.shutdown()

    async def _reap(self, client):
        # Ping client and allow time for a response
        # If a message is received in the meantime, this task is cancelled by the timer
        logging.info('Ping client {} pending reaping'.format(client.id))
        client.ping()

        await asyncio.sleep(self._reap_timeout)
        await self.remove_client(client)

    def get_room(self, room_id):
        return self.rooms.get(room_id)

    def get_public_rooms(self):
        return sorted([room for room in self.rooms.values() if room.is_public],
                      key=lambda room: room.active_ago)

    def create_room(self, name, **kwargs):
        room_id = slugify(name)
        if room_id in self.rooms:
            raise ChatException('Room {} already exists'.format(room_id))

        room = ChatRoom(name, **kwargs)
        self.add_room(room)

        @room.on('expire')
        async def on_expire():
            await self.remove_room(room)

        return room

    async def remove_room(self, room):
        self.rooms.pop(room.id, None)
        await room.shutdown()
        logging.info('Removed room %s', room.id)

    def create_client(self, client_id=None):
        if client_id is None:
            client_id = uuid.uuid4().hex

        if client_id in self.clients:
            raise ChatException('Client {} already exists'.format(client_id))

        client = ChatClient(client_id)
        client.timer = MTimer(self._reap_timeout, self._reap, client=client)

        greeting = ChatMessage()
        greeting.sender = self._message_address
        greeting.receiver = client.id
        greeting.type = 'greeting'
        greeting.data = 'This is Ground Control to Major Tom: You\'ve really made the grade. Now it\'s time to leave the capsule if you dare.'
        client.send(greeting.json())

        @client.on("message")
        async def on_message(message):
            # Reap this client if we haven't seen it for too long
            if client.timer is not None:
                client.timer.cancel()
            client.timer = MTimer(self._reap_timeout, self._reap, client=client)

            await self._handle_message(message, client)

        return client

    def get_ice_servers(self, client_id):
        stun_host = app.config['STUN_HOST']
        stun_port = app.config['STUN_PORT']
        stun_url = 'stun:{}:{}'.format(stun_host, stun_port)
        servers = [{'urls': [stun_url]}]

        turn_host = app.config['TURN_HOST']
        turn_port = app.config['TURN_PORT']
        turn_key = app.config['TURN_STATIC_AUTH_SECRET']

        if turn_host and turn_port and turn_key:
            turn_url = 'turn:{}:{}'.format(turn_host, turn_port)
            username, password = generate_turn_creds(turn_key, client_id)
            servers.append({'urls': [turn_url], 'username': username, 'credential': password})

        return servers



from setuptools import setup, find_packages

setup(
    name='camus-chat',
    version='0.1.0',
    description='Peer-to-peer video chat using WebRTC',
    long_description=open('README.rst').read(),
    long_description_content_type="text/x-rst; charset=UTF-8",
    url='https://camus.chat/',
    download_url='https://pypi.org/project/camus-chat/',
    project_urls={
        'Code': 'https://github.com/mrgnr/camus',
        'Issue tracker': 'https://github.com/mrgnr/camus/issues',
        'Documentation': 'https://docs.camus.chat/',
    },
    author='Morgan Robertson',
    author_email='mrgnr@pm.me',
    license='AGPL-3.0-or-later',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'quart==0.13.0',
        'bootstrap-flask==1.3.1',
        'flask-wtf==0.14.3',
        'pyee==7.0.2',
        'python-slugify==4.0.1',
        'werkzeug==1.0.1',
    ],
    classifiers=[
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Communications :: Conferencing',
        'Topic :: Multimedia :: Video',
    ],
    python_requires='>=3.7',
    scripts=['bin/camus'],
)



import os
basedir = os.path.abspath(os.path.dirname(__file__))

class Config(object):
    SECRET_KEY = os.environ.get('SECRET_KEY') or '420sixtyn9ne'
    STUN_HOST = os.environ.get('STUN_HOST') or 'stun.l.google.com'
    STUN_PORT = os.environ.get('STUN_PORT') or 19302
    TURN_HOST = os.environ.get('TURN_HOST')
    TURN_PORT = os.environ.get('TURN_PORT')
    TURN_STATIC_AUTH_SECRET = os.environ.get('TURN_STATIC_AUTH_SECRET')



import asyncio

from quart import (copy_current_websocket_context, flash, jsonify, redirect, render_template,
                   request, websocket)

from camus import app
from camus.forms import RoomCreate, RoomJoin

from camus import chat
from camus.chat import ChatException


@app.route('/')
@app.route('/index')
async def index():
    return redirect('/chat')

@app.route('/about')
async def about():
    return await render_template('about.html', title='Camus Video Chat | About')

@app.route('/chat', methods=['GET', 'POST'])
async def chat_create():
    manager = chat.get_chat_manager()

    form_create = RoomCreate()
    if form_create.validate_on_submit():
        form = form_create
        room_name = form.room_name.data
        password = None if not len(form.password.data) else form.password.data
        is_public = form.public.data
        guest_limit = None if form.guest_limit.data == 0 else form.guest_limit.data
        admin_list = []

        try:
            room = manager.create_room(room_name, password=password, guest_limit=guest_limit,
                                       admin_list=admin_list, is_public=is_public)
            return redirect('/chat/{}'.format(room.id))
        except ChatException as e:
            await flash('Room name not available')

    form_join = RoomJoin()
    public_rooms = manager.get_public_rooms()
    return await render_template('chat.html', form_create=form_create,
                                 form_join=form_join, public_rooms=public_rooms)


@app.route('/chat/<room_id>', methods=['GET', 'POST'])
async def chat_room(room_id):
    manager = chat.get_chat_manager()
    room = manager.get_room(room_id)

    if room is None:
        return '404', 404

    if room.is_full():
        return 'Guest limit already reached', 418

    if room.authenticate():  # i.e. a password is not required
        return await render_template('chatroom.html', title='Camus | {}'.format(room.name))

    form = RoomJoin()
    if form.validate_on_submit():
        room_id = form.room_id.data
        password = form.password.data

        if room.authenticate(password):
            # TODO: Generate token to be used with offer
            return await render_template('chatroom.html', title='Camus | {}'.format(room.name))
        else:
            await flash('Invalid password')

    return await render_template('join-room.html', title='Camus | Join a room', form=form, room_id=room_id)


@app.websocket('/chat/<room_id>/ws')
async def chat_room_ws(room_id):
    manager = chat.get_chat_manager()
    room = manager.get_room(room_id)

    if room is None:
        return # close the websocket

    client = manager.create_client()
    room.add_client(client)

    send_task = asyncio.create_task(
        copy_current_websocket_context(ws_send)(client.outbox),
    )
    receive_task = asyncio.create_task(
        copy_current_websocket_context(ws_receive)(client.inbox),
    )
    try:
        await asyncio.gather(send_task, receive_task)
    finally:
        send_task.cancel()
        receive_task.cancel()


async def ws_send(queue):
    while True:
        message = await queue.get()
        await websocket.send(message)


async def ws_receive(queue):
    while True:
        message = await websocket.receive()
        await queue.put(message)
